unit main;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, Menus, Plus4;

type
  TFormMain = class(TForm)
    MainMenu1: TMainMenu;
    OpenDialog1: TOpenDialog;
    Label1: TLabel;
		EditFileName: TEdit;
    cmdBrowse: TButton;
		cmdConvert: TButton;
		MemoDump: TMemo;
    Label2: TLabel;
    optFormatPlain: TRadioButton;
    optFormatPlus4IDE: TRadioButton;
    Label3: TLabel;
		procedure cmdBrowseClick(Sender: TObject);
		procedure cmdConvertClick(Sender: TObject);
		procedure FormCreate(Sender: TObject);
	private
		{ Private declarations }
	public
		{ Public declarations }
		Procedure ConvertFile ( s : String );
		Function PetChr ( b : Byte ) : Char;
	end;

var
	FormMain: TFormMain;

implementation

{$R *.DFM}

function TFormMain.PetChr ( b : Byte ) : Char;
begin
	If ( b >= $41 ) And ( b <= $5A ) Then
	begin
		PetChr := Chr ( b + $20 );
	End
	Else
	begin
		If ( b >= $C1 ) And ( b <= $DA ) Then
		begin
			PetChr := Chr ( b - $80 );
		End
		Else
		begin
			PetChr := Chr ( b );
		end;
	end;
End;

procedure TFormMain.cmdBrowseClick(Sender: TObject);
begin
	(*
	If OpenDialog1.InitialDir = '' Then
		OpenDialog1.InitialDir := 'C:\PROJECTS\PLUS4\Projects\!oldstuff\Part9-TheEnd\';
		*)
	//
	If OpenDialog1.Execute Then
	begin
		EditFileName.Text := OpenDialog1.FileName;
	End;
end;

procedure TFormMain.cmdConvertClick(Sender: TObject);
begin
	ConvertFile ( EditFileName.Text );
end;

Procedure TFormMain.ConvertFile ( s : String );
Var
	SourceFile : File;
	OutputFile : TextFile;

	d : Array [ 0 .. 64000 ] Of Byte;
	Labels : Array [ 0 .. 1023 ] Of String;
	nLabels : Integer;

	i : Integer;
	b : Byte;

	TabSize : Integer;
	iFileSize : Integer;
	Len : Integer;
	Flags : Integer;
	sTemp : String;
	OpCode : Byte;
	bNeedComma : Boolean;

	p : Integer; // pointer, increases through line info
	pData : Integer; // decrases through data
	p2 : Integer; // internal pointer within line

Procedure Dump ( sLine : String );
begin
	MemoDump.Lines.Add ( sLine );
	WriteLn ( OutputFile, sLine );
End;

Procedure Tab;
begin
	If optFormatPlus4IDE.Checked Then
		sTemp := sTemp + #9
	Else
		sTemp := sTemp + Copy ( '                ', 1, TabSize - Length ( sTemp ) );
End;

Procedure OpCodePostfix;
begin
	// close bracket BEFORE index
	If OpCodes [ OpCode ].ad = 9 Then
		sTemp := sTemp + ')';
	Case OpCodes [ OpCode ].ad Of
		3, 5, 10 :
		begin
			sTemp := sTemp + ',X';
		End;
		9, 7, 12 :
		begin
			sTemp := sTemp + ',Y';
		End;
	End;
	// close bracket AFTER index
	If ( OpCodes [ OpCode ].ad = 8 ) Or ( OpCodes [ OpCode ].ad = 10 ) Then
		sTemp := sTemp + ')';
	//
	OpCode := 0; // to prevent doing this twice
end;

begin
	AssignFile ( SourceFile, s );
	Reset ( SourceFile, 1 );
	//
	iFileSize := FileSize ( SourceFile );
	BlockRead ( SourceFile, d, iFileSize );
	CloseFile ( SourceFile );
	//
	AssignFile ( OutputFile, s + '.asm' );
	ReWrite ( OutputFile );
	//
	MemoDump.Visible := False;
	MemoDump.Lines.Clear;
	MemoDump.Lines.Add ( '; Loaded ' + IntToStr ( iFileSize ) + ' bytes' );
	//
	TabSize := d [ 14 ];
	//
	i := $100;
	p := 0;
	While ( i < iFileSize - 6 ) And ( p = 0 ) Do
	begin
		// look for TURBO marker
		If ( d[i+1]<>$54 ) Or ( d[i+2]<>$55 )
		Or ( d[i+3]<>$52 ) Or ( d[i+4]<>$42 ) Or ( d[i+5]<>$4F ) Then
		begin
			Inc ( i );
		End
		Else
		begin
			p := i; // found
		End;
	End;
	//
	If ( p = 0 ) Then
	begin
		MemoDump.Lines.Add ( '; TURBO marker not found. Probably not Turbo Assembler file.' );
	End
	Else
	begin
		MemoDump.Lines.Add ( '; TURBO marker found at $' + IntToHex ( p, 4 ) );
		pData := p;
		Inc ( p, 6 );
		While ( d [ p ] <> $60 ) And ( p < iFileSize ) Do
		begin
			Inc ( p );
		End;
		Inc ( p, $60 );
		MemoDump.Lines.Add ( '; LABELS from $' + IntToHex ( p, 4 ) );
		//
		// ---
		//
		i := 0; // label counter
		While ( d [ p ] <> $FF ) And ( p < iFileSize ) Do
		begin
			b := d [ p ];
			Labels [ i ] := Labels [ i ] + PetChr ( b And $7F );
			Inc ( p );
			//
			If ( b And $80 ) <> 0 Then
			begin
				//Dump ( IntToHex ( i, 4 ) + Labels [ i ] );
				Inc ( i );
			End;
		End;
		nLabels := i - 1;
		//
		// ---
		//
		p := pData + 6; // rewind
		//
		While d [ p ] <> $60 Do
		begin
			Len := d [ p ] And $3F;
			Flags := d [ p ] And $C0;
			//
			Dec ( pData, Len );
			//
			p2 := 0;
			OpCode := 0;
			sTemp := '';
			//
			If Len > 0 Then
			begin
				{
				sTemp := '[$' + IntToHex ( Flags, 2 ) + ']';
				For i := 0 To Len - 1 Do
				begin
					sTemp := sTemp + ' ' + IntToHex ( d [ pData + i ], 2 );
				End;
				}
				//
				If Flags = $C0 Then
				begin
					// this line has a label
					sTemp := sTemp + Labels [ ( d [ pData + p2 ] And 7 ) * 256 + d [ pData + p2 + 1 ] ];
					Inc ( p2, 2 );
				end;
				//
				If ( Flags And $80 ) = $80 Then
				begin
					//
					Tab;
					//
					OpCode := d [ pData + p2 ];
					sTemp := sTemp + OpCodes [ OpCode ].mn + ' ';
					//
					Case OpCodes [ OpCode ].ad Of
						2 :
						begin
							sTemp := sTemp + '#';
						End;
						8, 9, 10 :
						begin
							sTemp := sTemp + '(';
						End;
						15 :
						Begin
							sTemp := sTemp + 'A'; // e.g. LSR A, etc.
						End;
					End;
					//
					Inc ( p2 );
				End;
				//
				// *** Process line
				//
				bNeedComma := False;
				While p2 < Len Do
				begin
					Case d [ pData + p2 ] Of
						$02 :
						begin
							Tab;
							sTemp := sTemp + '.text "';
							Inc ( p2 );
							//
							For i := 1 To d [ pData + p2 ] Do
							begin
								sTemp := sTemp + PetChr ( d [ pData + p2 + i ] );
							End;
							sTemp := sTemp + '"';
							Inc ( p2, d [ pData + p2 ] + 1 );
							bNeedComma := False;
						End;
						$03 :
						begin
							Tab;
							If optFormatPlain.Checked Then
								sTemp := sTemp + '.byte '
							Else
								sTemp := sTemp + 'DB ';
							Inc ( p2 );
							bNeedComma := False;
						End;
						$04 :
						begin
							Tab;
							If optFormatPlain.Checked Then
								sTemp := sTemp + '.word '
							Else
								sTemp := sTemp + 'DW ';
							Inc ( p2 );
							bNeedComma := False;
						End;
						$05 :
						begin
							Tab;
							sTemp := sTemp + '= ';
							Inc ( p2 );
							bNeedComma := False;
						End;
						$06 :
						begin
							Tab;
							If optFormatPlain.Checked Then
								sTemp := sTemp + '*= '
							Else
								sTemp := sTemp + 'ORG ';
							Inc ( p2 );
							bNeedComma := False;
						End;
						$20,$28 :
						begin
							// 8 bit hex
							If bNeedComma Then sTemp := sTemp + ',';
							sTemp := sTemp + '$';
							sTemp := sTemp + IntToHex ( d [ pData + p2 + 1 ], 2 );
							Inc ( p2, 2 );
							bNeedComma := True;
						End;
						$21,$29 :
						begin
							// 16 bit hex
							If bNeedComma Then sTemp := sTemp + ',';
							sTemp := sTemp + '$';
							sTemp := sTemp + IntToHex ( d [ pData + p2 + 2 ], 2 ) + IntToHex ( d [ pData + p2 + 1 ], 2 );
							Inc ( p2, 3 );
							bNeedComma := True;
						End;
						$22,$2A :
						begin
							// 8 bit dec
							If bNeedComma Then sTemp := sTemp + ',';
							sTemp := sTemp + IntToStr ( d [ pData + p2 + 1 ] );
							Inc ( p2, 2 );
							bNeedComma := True;
						End;
						$23,$2B :
						begin
							// 16 bit dec
							If bNeedComma Then sTemp := sTemp + ',';
							sTemp := sTemp + IntToStr ( d [ pData + p2 + 2 ] * 256 + d [ pData + p2 + 1 ] );
							Inc ( p2, 3 );
							bNeedComma := True;
						End;
						$26,$2E :
						begin
							// 8 bit char
							If bNeedComma Then sTemp := sTemp + ',';
							sTemp := sTemp + '"' + PetChr ( d [ pData + p2 + 1 ] ) + '"';
							Inc ( p2, 2 );
							bNeedComma := True;
						End;
						$30,$31,$32,$33,$34,$35,$36,$37,
						$38,$39,$3A,$3B,$3C,$3D,$3E,$3F :
						begin
							i := ( d [ pData + p2 ] And 7 ) * 256 + d [ pData + p2 + 1 ];
							If ( i <= nLabels ) Then
								sTemp := sTemp + Labels [ i ];
							Inc ( p2, 2 );
							bNeedComma := True;
						End;
						$40 :
						begin
							sTemp := sTemp + '+';
							Inc ( p2, 1 );
							bNeedComma := False;
						End;
						$41 :
						begin
							sTemp := sTemp + '-';
							Inc ( p2, 1 );
							bNeedComma := False;
						End;
						$42 :
						begin
							sTemp := sTemp + '*';
							Inc ( p2, 1 );
							bNeedComma := False;
						End;
						$43 :
						begin
							sTemp := sTemp + '/';
							Inc ( p2, 1 );
							bNeedComma := False;
						End;
						$44 :
						begin
							sTemp := sTemp + '>';
							Inc ( p2, 1 );
							bNeedComma := False;
						End;
						$45 :
						begin
							sTemp := sTemp + '<';
							Inc ( p2, 1 );
							bNeedComma := False;
						End;
						$80,$81,$82,$83,$84,$85,$86,$87,
						$88,$89,$8A,$8B,$8C,$8D,$8E,$8F,
						$90,$91,$92,$93,$94,$95,$96,$97,
						$98,$99,$9A,$9B,$9C,$9D,$9E,$9F,
						$A0,$A1,$A2,$A3,$A4,$A5,$A6,$A7 :
						begin
							//
							// comment
							//
							OpCodePostfix;
							//
							b := d [ pData + p2 ] - $80;
							If b > Length ( sTemp ) Then
								sTemp := sTemp + Copy ( '                ', 1, b - Length ( sTemp ) );
							sTemp := sTemp + ';';
							//
							Inc ( p2 );
							//
							While ( p2 < Len ) Do
							begin
								b := d [ pData + p2 ];
								sTemp := sTemp + PetChr ( b );
								Inc ( p2 );
							End;
							bNeedComma := False;
						End;
						Else
						begin
							// ShowMessage ( 'Unhandled: ' + IntTOHex ( d [pData+p2],2 ) );
							Dump ( '******' + 'Unhandled: ' + IntTOHex ( d [pData+p2],2 ) );
							Inc ( p2 );
						End;
					End;
				End;
				//
				OpCodePostfix;
			End;
			//
			Dump ( sTemp );
			//
			Inc ( p );
		End;
	End;
	//
	CloseFile ( OutputFile );
	//
	MemoDump.Visible := True;
end;

procedure TFormMain.FormCreate(Sender: TObject);
begin
	//EditFileName.Text := 'C:\PROJECTS\PLUS4\Projects\!oldstuff\Part9-Theend\THE_end\four.prg';
	//EditFileName.Text := 'C:\PROJECTS\PLUS4\Projects\!oldstuff\Part9-Theend\TE V20.10.prg';
end;

end.
